import logging
from datetime import datetime
from typing import Optional, Iterable

from meross_iot.controller.device import GenericSubDevice
from meross_iot.model.enums import Namespace, OnlineStatus, ThermostatV3Mode

_LOGGER = logging.getLogger(__name__)


class Ms100Sensor(GenericSubDevice):
    """
    This class maps the functionality offered by the MS100 sensor device.
    The MS100 offers temperature and humidity sensing.
    Moreover, this device is capable of triggering settable alerts.
    """
    _UPDATE_ALL_NAMESPACE = Namespace.HUB_SENSOR_ALL

    def __init__(self, hubdevice_uuid: str, subdevice_id: str, manager, **kwargs):
        super().__init__(hubdevice_uuid, subdevice_id, manager, **kwargs)
        self.__temperature = {}
        self.__humidity = {}
        self.__samples = []

    async def _execute_command(self, method: str, namespace: Namespace, payload: dict, timeout: Optional[float] = None) -> dict:
        raise NotImplementedError("This method should never be called directly for subdevices.")

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False
        if namespace == Namespace.HUB_ONLINE:
            update_element = self._prepare_push_notification_data(data=data, filter_accessor='online')
            if update_element is not None:
                self._online = OnlineStatus(update_element.get('status', -1))
                locally_handled = True
        return locally_handled

    async def async_handle_subdevice_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False
        if namespace == Namespace.HUB_ONLINE:
            self._online = OnlineStatus(data.get('online', {}).get('status', -1))
            self._last_active_time = data.get('online', {}).get('lastActiveTime')
        elif namespace == Namespace.HUB_SENSOR_ALL:
            self._online = OnlineStatus(data.get('online', {}).get('status', -1))
            self.__temperature.update(data.get('temperature', {}))
            self.__humidity.update(data.get('humidity', {}))
            locally_handled = True
        elif namespace == Namespace.HUB_SENSOR_TEMPHUM:
            latest_temperature = data.get('latestTemperature')
            latest_humidity = data.get('latestHumidity')
            synced_time = data.get('syncedTime')
            samples = data.get('sample')
            if synced_time is not None and (self.last_sampled_time is None or
                                            synced_time > self.last_sampled_time.timestamp()):
                self.__temperature['latestSampleTime'] = synced_time
                self.__temperature['latest'] = latest_temperature
                self.__humidity['latestSampleTime'] = synced_time
                self.__humidity['latest'] = latest_humidity

            self.__samples.clear()
            for sample in samples:
                temp, hum, from_ts, to_ts, unknown = sample
                self.__samples.append({
                    'from_ts': from_ts,
                    'to_ts': to_ts,
                    'temperature': float(temp) / 10,
                    'humidity': float(hum) / 10
                })

            else:
                _LOGGER.debug("Skipping temperature update as synched time is None or old compared to the latest data")
            locally_handled = True
        elif namespace == Namespace.HUB_SENSOR_ALERT:
            locally_handled = False
            # TODO: not yet implemented
        else:
            _LOGGER.warning(f"Could not handle event %s in subdevice %s handler", namespace, self.name)

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    @property
    def last_sampled_temperature(self) -> Optional[float]:
        """
        Returns the latest sampled temperature in Celsius degrees.
        If you want to refresh this data, call `async_update` to force a full
        data refresh.

        :return: The latest sampled temperature, if available, in Celsius degree
        """
        temp = self.__temperature.get('latest')
        if temp is None:
            return None
        return float(temp) / 10.0

    @property
    def last_sampled_humidity(self) -> Optional[float]:
        """
        Exposes the latest sampled humidity, in %.
        If you want to refresh this data, call `async_update` to force a full
        data refresh.

        :return: The latest sampled humidity grade in %, if available
        """
        humidity = self.__humidity.get('latest')
        if humidity is None:
            return None
        return float(humidity) / 10.0

    @property
    def last_sampled_time(self) -> Optional[datetime]:
        """
        UTC datetime when the latest update has been sampled by the sensor

        :return: latest sampling time in UTC, if available
        """
        timestamp = self.__temperature.get('latestSampleTime')
        if timestamp is None:
            return None

        return datetime.utcfromtimestamp(timestamp)

    @property
    def min_supported_temperature(self) -> Optional[float]:
        """
        Maximum supported temperature that this device can report

        :return: float value, maximum supported temperature, if available
        """
        return self.__temperature.get('min')

    @property
    def max_supported_temperature(self) -> Optional[float]:
        """
        Minimum supported temperature that this device can report
        """
        return self.__temperature.get('max')


class Mts100v3Valve(GenericSubDevice):
    _UPDATE_ALL_NAMESPACE = Namespace.HUB_MTS100_ALL

    def __init__(self, hubdevice_uuid: str, subdevice_id: str, manager, **kwargs):
        super().__init__(hubdevice_uuid, subdevice_id, manager, **kwargs)
        self.__togglex = {}
        self.__timeSync = None
        self.__mode = {}
        self.__temperature = {}
        self._schedule_b_mode = None
        self._last_active_time = None
        self.__adjust = {}

    async def _execute_command(self, method: str, namespace: Namespace, payload: dict, timeout: Optional[float] = None) -> dict:
        raise NotImplementedError("This method should never be called directly for subdevices.")

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False
        if namespace == Namespace.HUB_ONLINE:
            update_element = self._prepare_push_notification_data(data=data, filter_accessor='online')
            if update_element is not None:
                self._online = OnlineStatus(update_element.get('status', -1))
                locally_handled = True
        return locally_handled

    async def async_handle_subdevice_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False
        if namespace == Namespace.HUB_ONLINE:
            self._online = OnlineStatus(data.get('online', {}).get('status', -1))
            self._last_active_time = data.get('online', {}).get('lastActiveTime')
        elif namespace == Namespace.HUB_MTS100_ALL:
            self._schedule_b_mode = data.get('scheduleBMode')
            self._online = OnlineStatus(data.get('online', {}).get('status', -1))
            self._last_active_time = data.get('online', {}).get('lastActiveTime')
            self.__togglex.update(data.get('togglex', {}))
            self.__timeSync = data.get('timeSync', {})
            self.__mode.update(data.get('mode', {}))
            self.__temperature.update(data.get('temperature', {}))
            self.__temperature['latestSampleTime'] = datetime.utcnow().timestamp()
            self.__adjust.update(data.get('temperature', {}))
            self.__adjust['latestSampleTime'] = datetime.utcnow().timestamp()
            locally_handled = True
        elif namespace == Namespace.HUB_TOGGLEX:
            update_element = self._prepare_push_notification_data(data=data)
            if update_element is not None:
                self.__togglex.update(update_element)
                locally_handled = True
        elif namespace == Namespace.HUB_MTS100_MODE:
            update_element = self._prepare_push_notification_data(data=data)
            if update_element is not None:
                self.__mode.update(update_element)
                locally_handled = True
        elif namespace == Namespace.HUB_MTS100_TEMPERATURE:
            update_element = self._prepare_push_notification_data(data=data)
            if update_element is not None:
                self.__temperature.update(update_element)
                self.__temperature['latestSampleTime'] = datetime.utcnow().timestamp()
                locally_handled = True
        else:
            _LOGGER.error(f"Could not handle event %s in subdevice %s handler", namespace, self.name)

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    def is_on(self) -> Optional[bool]:
        return self.__togglex.get('onoff') == 1

    async def async_turn_off(self, timeout: Optional[float] = None, *args, **kwargs):
        await self._hub._execute_command(method="SET", namespace=Namespace.HUB_TOGGLEX,
                                         payload={'togglex': [{"id": self.subdevice_id, "onoff": 0, "channel": 0}]},
                                         timeout=timeout)
        # Assume the command was ok, so immediately update the internal state
        self.__togglex['onoff'] = 0

    async def async_turn_on(self, timeout: Optional[float] = None,
                            *args, **kwargs):
        await self._hub._execute_command(method="SET", namespace=Namespace.HUB_TOGGLEX,
                                         payload={'togglex': [{"id": self.subdevice_id, "onoff": 1, "channel": 0}]},
                                         timeout=timeout)
        # Assume the command was ok, so immediately update the internal state
        self.__togglex['onoff'] = 1

    async def async_toggle(self, timeout: float = None, *args, **kwargs):
        if self.is_on():
            await self.async_turn_off(timeout=timeout)
        else:
            await self.async_turn_on(timeout=timeout)

    @property
    def last_sampled_temperature(self) -> Optional[float]:
        """
        Current room temperature in Celsius degrees.

        :return: float number
        """
        temp = self.__temperature.get('room')
        if temp is not None:
            return float(temp) / 10.0
        else:
            return None

    async def async_get_temperature(self, timeout: Optional[float] = None, *args, **kwargs) -> Optional[float]:
        """
        Polls the device in order to retrieve the latest temperature info.
        You should not use this method so ofter: instead, rely on `last_sampled_temperature` when a cached
        value is ok.

        :return:
        """
        res = await self._hub._execute_command(method="GET", namespace=Namespace.HUB_MTS100_TEMPERATURE,
                                               payload={'temperature': [{"id": self.subdevice_id}]}, timeout=timeout)
        if res is None:
            return None

        for d in res.get('temperature'):
            if d.get('id') == self.subdevice_id:
                del d['id']
                self.__temperature.update(d)
                self.__temperature['latestSampleTime'] = datetime.utcnow().timestamp()
                break

        return self.last_sampled_temperature

    @property
    def last_sampled_time(self) -> Optional[datetime]:
        """
        UTC datetime when the latest update has been sampled by the sensor

        :return: latest sampling time in UTC, if available
        """
        timestamp = self.__temperature.get('latestSampleTime')
        if timestamp is None:
            return None

        return datetime.fromtimestamp(timestamp)

    @property
    def mode(self) -> Optional[ThermostatV3Mode]:
        m = self.__mode.get('state')
        if m is not None:
            return ThermostatV3Mode(m)

    async def async_set_mode(self, mode: ThermostatV3Mode, timeout: Optional[float] = None, *args, **kwargs) -> None:
        payload = {'mode': [{'id': self.subdevice_id, 'state': mode.value}]}
        await self._hub._execute_command(method='SET', namespace=Namespace.HUB_MTS100_MODE, payload=payload,
                                         timeout=timeout)
        self.__mode['state'] = mode.value

    @property
    def target_temperature(self) -> Optional[float]:
        temp = self.__temperature.get('currentSet')
        if temp is not None:
            return float(temp) / 10.0
        else:
            return None

    @property
    def min_supported_temperature(self) -> Optional[float]:
        temp = self.__temperature.get('min')
        if temp is not None:
            return float(temp) / 10.0
        else:
            return None

    @property
    def max_supported_temperature(self) -> Optional[float]:
        temp = self.__temperature.get('max')
        if temp is not None:
            return float(temp) / 10.0
        else:
            return None

    @property
    def is_heating(self) -> Optional[bool]:
        return self.__temperature.get('heating') == 1

    @property
    def is_window_open(self) -> Optional[bool]:
        return self.__temperature.get('openWindow') == 1

    def get_preset_temperature(self, preset: str) -> Optional[float]:
        """
        Returns the current set temperature for the given preset.

        :param preset:

        :return: float temperature value
        """
        if preset not in self.get_supported_presets():
            _LOGGER.error(f"Preset {preset} is not supported by this device.")
        val = self.__temperature.get(preset)
        if val is None:
            return None
        return float(val) / 10.0

    @staticmethod
    def get_supported_presets() -> Iterable[str]:
        """
        Returns the supported presets of this device.

        :return: an iterable of strings
        """
        return 'custom', 'comfort', 'economy', 'away'

    async def async_set_preset_temperature(self, preset: str, temperature: float, timeout: Optional[float] = None, *args,
                                           **kwargs) -> None:
        """
        Sets the preset temperature configuration.

        :param preset: string preset, as reported by `get_supported_presets()`
        :param temperature: temperature to be set for the given preset

        :return: None
        """
        if preset not in self.get_supported_presets():
            raise ValueError(f"Preset {preset} is not supported by this device. "
                             f"Valid presets are: {self.get_supported_presets()}")
        target_temp = temperature * 10
        await self._hub._execute_command(method="SET", namespace=Namespace.HUB_MTS100_TEMPERATURE,
                                         payload={'temperature': [{'id': self.subdevice_id, preset: target_temp}]},
                                         timeout=timeout)

        # Update local state
        self.__temperature[preset] = target_temp

    async def async_set_target_temperature(self, temperature: float, timeout: Optional[float] = None, *args,
                                           **kwargs) -> None:
        # The API expects the target temperature in DECIMALS, so we need to multiply the user's input by 10
        target_temp = temperature * 10
        payload = {'temperature': [{'id': self.subdevice_id, 'custom': target_temp}]}
        await self._hub._execute_command(method='SET', namespace=Namespace.HUB_MTS100_TEMPERATURE, payload=payload,
                                         timeout=timeout)
        # Update local state
        self.__temperature['currentSet'] = target_temp

    async def async_get_adjust(self, timeout: Optional[float] = None, *args, **kwargs) -> Optional[float]:
        """
        :return:
        """
        res = await self._hub._execute_command(method="GET", namespace=Namespace.HUB_MTS100_ADJUST,
                                               payload={'adjust': [{"id": self.subdevice_id}]}, timeout=timeout)
        if res is None:
            return None

        for d in res.get('adjust'):
            if d.get('id') == self.subdevice_id:
                del d['id']
                self.__adjust.update(d)
                self.__adjust['latestSampleTime'] = datetime.utcnow().timestamp()
                break

        return self.adjust

    @property
    def adjust(self) -> Optional[float]:
        """
        Returns the adjust temperature value for the sensor if available

        :return:
        """
        adjust = self.__adjust.get('temperature')
        if adjust is None:
            return None

        return float(self.__adjust.get('temperature')) / 100.0

    async def async_set_adjust(self, temperature: float, timeout: Optional[float] = None) -> None:
        # The API expects the adjust temperature in HUNDREDS (not consistent with the temperature set), so we need to multiply the user's input by 100
        # N.B. the App enforces on the frontend a limit on the adjustment (+/- 5 C°), tests show there is no limit on the API
        adjust_temp = temperature * 100
        payload = {'adjust': [{'id': self.subdevice_id, 'temperature': adjust_temp}]}
        await self._hub._execute_command(method='SET', namespace=Namespace.HUB_MTS100_ADJUST, payload=payload,
                                         timeout=timeout)
        # Update local state
        self.__adjust.update({'temperature': adjust_temp})
        self.__adjust['latestSampleTime'] = datetime.utcnow().timestamp()



import logging
from typing import Optional

_LOGGER = logging.getLogger(__name__)


def extract_domain(address: str) -> str:
    tokens = address.split(":")
    return tokens[0]

def extract_port(address: str, default: int) -> int:
    tokens = address.split(":")
    if len(tokens) > 1:
        return tokens[1]
    return default




from datetime import datetime


class PowerInfo(object):
    def __init__(self, current_ampere: float, voltage_volts: float, power_watts: float, sample_timestamp: datetime):
        self._current = current_ampere
        self._voltage = voltage_volts
        self._power = power_watts
        self._sample_timestamp = sample_timestamp

    @property
    def power(self) -> float:
        return self._power

    @property
    def voltage(self) -> float:
        return self._voltage

    @property
    def current(self) -> float:
        return self._current

    @property
    def sample_timestamp(self) -> datetime:
        return self._sample_timestamp

    def __str__(self):
        return f"POWER = {self._power} W, VOLTAGE = {self._voltage} V, CURRENT = {self._current} A"



import logging
from typing import Optional

from meross_iot.model.enums import Namespace, OnlineStatus

_LOGGER = logging.getLogger(__name__)


class SystemAllMixin(object):
    _execute_command: callable
    #async_handle_update: Callable[[Namespace, dict], Awaitable]

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)

    async def async_update(self, timeout: Optional[float] = None, *args, **kwargs) -> None:
        # Call the super implementation
        await super().async_update(timeout=timeout, *args, **kwargs)

        result = await self._execute_command(method="GET",
                                             namespace=Namespace.SYSTEM_ALL,
                                             payload={},
                                             timeout=timeout)

        # Once we have the response, update all the mixin which are interested
        await self.async_handle_update(namespace=Namespace.SYSTEM_ALL, data=result)


class SystemOnlineMixin(object):
    _online: OnlineStatus
    #async_handle_update: Callable[[Namespace, dict], Awaitable]

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)

    async def async_handle_update(self, namespace: Namespace, data: dict) -> bool:
        _LOGGER.debug(f"Handling {self.__class__.__name__} mixin data update.")
        locally_handled = False
        if namespace == Namespace.SYSTEM_ALL:
            online_data = data.get('all').get('system').get('online')
            status = OnlineStatus(int(online_data.get("status")))
            self._online = status
            locally_handled = True

        super_handled = await super().async_handle_update(namespace=namespace, data=data)
        return super_handled or locally_handled

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False

        if namespace == Namespace.SYSTEM_ONLINE:
            _LOGGER.debug(f"OnlineMixin handling push notification for namespace {namespace}")
            payload = data.get('online')
            if payload is None:
                _LOGGER.error(f"OnlineMixin could not find 'online' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                status = OnlineStatus(int(payload.get("status")))
                self._online = status
                locally_handled = True

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled



import os

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.mixins.toggle import ToggleMixin
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus
from tests import async_get_client

if os.name == 'nt':
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestToggle(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        devices = await self.meross_manager.async_device_discovery()
        toggle_devices = self.meross_manager.find_devices(device_class=ToggleMixin, online_status=OnlineStatus.ONLINE)

        if len(toggle_devices) < 1:
            self.test_device = None
        else:
            self.test_device = toggle_devices[0]

    @unittest_run_loop
    async def test_toggle_local_state(self):
        if self.test_device is None:
            self.skipTest("No ToggleX device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")
        # Turn off device to start from a clean state
        r = await self.test_device.async_turn_off()
        self.assertFalse(self.test_device.is_on())
        await asyncio.sleep(1.0)

        # Turn on the device
        r = await self.test_device.async_turn_on()
        self.assertTrue(self.test_device.is_on())

        # Turn off the device
        await asyncio.sleep(1)
        r = await self.test_device.async_turn_off()
        self.assertFalse(self.test_device.is_on())

    @unittest_run_loop
    async def test_toggle_push_notification(self):
        if self.test_device is None:
            self.skipTest("No ToggleX device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")
        # Create a new manager
        new_meross_client, requires_logout = await async_get_client()
        m = None
        try:
            # Retrieve the same device with another manager
            m = MerossManager(http_client=new_meross_client)
            await m.async_init()
            await m.async_device_discovery()
            devs = m.find_devices(device_uuids=(self.test_device.uuid,))
            dev = devs[0]

            # Turn off device to start from a clean state
            r = await self.test_device.async_turn_off()
            await asyncio.sleep(2)

            # Turn on the device
            r = await self.test_device.async_turn_on()
            # Wait a bit and make sure the other manager received the push notification
            await asyncio.sleep(2)
            self.assertTrue(self.test_device.is_on())
            self.assertTrue(dev.is_on())

            # Turn off the device
            await asyncio.sleep(1)
            r = await self.test_device.async_turn_off()
            # Wait a bit and make sure the other manager received the push notification
            await asyncio.sleep(2)
            self.assertFalse(self.test_device.is_on())
            self.assertFalse(dev.is_on())

        finally:
            if m is not None:
                m.close()
            if requires_logout:
                await new_meross_client.async_logout()

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)


import concurrent
import os
from random import randint

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.device import BaseDevice
from meross_iot.controller.mixins.light import LightMixin
from meross_iot.http_api import MerossHttpClient
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus
from tests import async_get_client


if os.name == 'nt':
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestLight(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        devices = await self.meross_manager.async_device_discovery()
        self.light_devices = self.meross_manager.find_devices(device_class=LightMixin, online_status=OnlineStatus.ONLINE)

        # Update the states of all devices a first time
        concurrent_update = [d.async_update() for d in self.light_devices]
        await asyncio.gather(*concurrent_update)

    @unittest_run_loop
    async def test_rgb(self):
        # Make sure we have an RGB-capable available device
        rgb_capable = [d for d in self.light_devices if d.get_supports_rgb()]
        if len(rgb_capable) < 1:
            self.skipTest("Could not find any RGB-capable LightMixin within the given set of devices. "
                          "The test will be skipped")
        for light in rgb_capable:
            print(f"Testing device: {light.name}...")
            await light.async_update()

            # Set a random color
            r = randint(0, 256)
            g = randint(0, 256)
            b = randint(0, 256)
            await light.async_set_light_color(rgb=(r, g, b), onoff=True)

            # Check the color property returns red
            color = light.get_rgb_color()
            self.assertEqual(color, (r, g, b))

    @unittest_run_loop
    async def test_turn_on_off(self):
        # Make sure we have an RGB-capable available device
        rgb_capable = [d for d in self.light_devices if d.get_supports_rgb()]
        if len(rgb_capable) < 1:
            self.skipTest("Could not find any RGB-capable LightMixin within the given set of devices. "
                          "The test will be skipped")
        for light in rgb_capable:
            print(f"Testing device: {light.name}...")
            await light.async_update()

            # Turn device off
            await light.async_set_light_color(rgb=(255, 255, 255), onoff=False)
            await asyncio.sleep(1)
            # Make sure device is now off
            self.assertFalse(light.get_light_is_on())

            # Set a color and turn the device on
            await light.async_set_light_color(rgb=(0, 255, 0), onoff=True)
            await asyncio.sleep(1)
            # Make sure device is now on with that specific color set
            self.assertTrue(light.get_light_is_on())
            self.assertEqual(light.get_rgb_color(), (0, 255, 0))

            # Set a color without changing the on-off state
            await light.async_set_light_color(rgb=(255, 0, 0))
            await asyncio.sleep(1)
            # Make sure device is now showing the specific color
            self.assertTrue(light.get_light_is_on())
            self.assertEqual(light.get_rgb_color(), (255, 0, 0))

            # Turn off device without changing color
            await light.async_set_light_color(onoff=False)
            await asyncio.sleep(1)
            # Make sure device is now off
            self.assertFalse(light.get_light_is_on())

    @unittest_run_loop
    async def test_rgb_push_notification(self):
        # Make sure we have an RGB-capable available device
        rgb_capable = list(filter(lambda d: d.get_supports_rgb(), self.light_devices))
        if len(rgb_capable) < 1:
            self.skipTest("Could not find any RGB-capable LightMixin within the given set of devices. "
                          "The test will be skipped")

        light: BaseDevice = rgb_capable[0]
        print(f"Selected test device: {light.name}.")

        # Create a new manager
        new_meross_client, requires_logout = await async_get_client()
        m = None
        try:
            # Retrieve the same device with another manager
            m = MerossManager(http_client=new_meross_client)
            await m.async_init()
            await m.async_device_discovery()
            devs = m.find_devices(device_uuids=(light.uuid,))
            if len(devs) < 1:
                self.skipTest("Could not find dev for push notification")
                return
            dev = devs[0]

            # Set RGB color to known state
            r = await light.async_set_light_color(rgb=(255, 0, 0), onoff=True)
            await asyncio.sleep(2)

            # Turn on the device
            r = await light.async_set_light_color(rgb=(0, 255, 0), onoff=True)

            # Wait a bit and make sure the other manager received the push notification
            await asyncio.sleep(10)
            self.assertEqual(light.get_rgb_color(), (0, 255, 0))
            self.assertEqual(dev.get_rgb_color(), (0, 255, 0))
        finally:
            if m is not None:
                m.close()
            if requires_logout:
                await new_meross_client.async_logout()

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)



from meross_iot.model.enums import Namespace, OnlineStatus
from meross_iot.model.push.generic import GenericPushNotification


class OnlinePushNotification(GenericPushNotification):
    def __init__(self, originating_device_uuid: str, raw_data: dict):
        super().__init__(namespace=Namespace.SYSTEM_ONLINE,
                         originating_device_uuid=originating_device_uuid,
                         raw_data=raw_data)

    @property
    def status(self) -> OnlineStatus:
        return self.raw_data.get('online', {}).get('status', None)



import os
from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.device import HubDevice
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus
from tests import async_get_client

if os.name == 'nt':
    import asyncio

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestHub(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        await self.meross_manager.async_device_discovery()
        self.test_devices = self.meross_manager.find_devices(device_class=HubDevice,
                                                             online_status=OnlineStatus.ONLINE)

    @unittest_run_loop
    async def test_update(self):
        if len(self.test_devices) < 1:
            self.skipTest("No HUB device has been found to run this test.")

        dev = self.test_devices[0]
        print(f"Testing device {dev.name}")
        await dev.async_update()

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)



import os

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.mixins.dnd import SystemDndMixin
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus, DNDMode
from tests import async_get_client

if os.name == 'nt':
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestDnd(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        await self.meross_manager.async_device_discovery()
        self.test_devices = self.meross_manager.find_devices(device_class=SystemDndMixin, online_status=OnlineStatus.ONLINE)
        print(f"Test devices: {self.test_devices}")

    @unittest_run_loop
    async def test_set_dnd_mode(self):
        if len(self.test_devices) < 1:
            self.skipTest("No device has been found to run this test UPDATE ALL on it.")

        # Toggle DNDMode
        for d in self.test_devices:
            print(f"Testing device {d.name}")
            print(f"Executing update on device {d.name} ({d.type})")
            await d.async_update()

            # Enable DND for every device
            await d.set_dnd_mode(DNDMode.DND_ENABLED)
            self.assertEqual(await d.async_get_dnd_mode(), DNDMode.DND_ENABLED)

            # Disable DND mode back
            await d.set_dnd_mode(DNDMode.DND_DISABLED)
            self.assertEqual(await d.async_get_dnd_mode(), DNDMode.DND_DISABLED)


    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)


import os
import random
from random import randint
from typing import Union

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.device import BaseDevice
from meross_iot.controller.mixins.diffuser_light import DiffuserLightMixin
from meross_iot.controller.mixins.diffuser_spray import DiffuserSprayMixin
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus, DiffuserLightMode, DiffuserSprayMode
from tests import async_get_client

if os.name == 'nt':
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestDiffuser(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        devices = await self.meross_manager.async_device_discovery()
        self.light_devices = self.meross_manager.find_devices(device_class=DiffuserLightMixin, online_status=OnlineStatus.ONLINE)
        self.spray_devices = self.meross_manager.find_devices(device_class=DiffuserSprayMixin, online_status=OnlineStatus.ONLINE)


    @unittest_run_loop
    async def test_light_rgb(self):
        if len(self.light_devices) < 1:
            self.skipTest("Could not find any DiffuserLightMixin within the given set of devices. "
                          "The test will be skipped.")
        for light in self.light_devices:
            print(f"Testing device: {light.name}...")
            await light.async_update()

            await light.async_set_light_mode(mode=DiffuserLightMode.FIXED_RGB)

            # Set a random color
            r = randint(0, 256)
            g = randint(0, 256)
            b = randint(0, 256)
            await light.async_set_light_mode(rgb=(r, g, b), onoff=True)

            # Check the color property returns red
            color = light.get_light_rgb_color()
            self.assertEqual(color, (r, g, b))

    @unittest_run_loop
    async def test_turn_light_on_off(self):
        if len(self.light_devices) < 1:
            self.skipTest("Could not find any DiffuserLightMixin within the given set of devices. "
                          "The test will be skipped.")
        for light in self.light_devices:
            print(f"Testing device: {light.name}...")
            await light.async_update()  # type: Union[BaseDevice, DiffuserLightMode]

            # Set mode to FIXED RGB
            await light.async_set_light_mode(mode=DiffuserLightMode.FIXED_RGB)

            # Turn device off
            await light.async_set_light_mode(onoff=False)
            self.assertFalse(light.get_light_is_on())
            # Change the color without turning it on
            await light.async_set_light_mode(rgb=(255, 0, 0))
            # Make sure device is now showing the specific color
            self.assertEqual(light.get_light_rgb_color(), (255, 0, 0))
            self.assertFalse(light.get_light_is_on())
            # Turn device on and check if the color is right
            await light.async_set_light_mode(onoff=True)
            self.assertTrue(light.get_light_is_on())

            # Turn off device without changing color
            await light.async_set_light_mode(onoff=False)
            await asyncio.sleep(1)
            # Make sure device is now off
            self.assertFalse(light.get_light_is_on())

    @unittest_run_loop
    async def test_light_mode(self):
        if len(self.light_devices) < 1:
            self.skipTest("Could not find any DiffuserLightMixin within the given set of devices. "
                          "The test will be skipped.")
        for light in self.light_devices:
            print(f"Testing device: {light.name}...")
            await light.async_update()

            await light.async_set_light_mode(mode=DiffuserLightMode.FIXED_LUMINANCE, onoff=True)
            self.assertEqual(light.get_light_mode(), DiffuserLightMode.FIXED_LUMINANCE)
            await light.async_set_light_mode(mode=DiffuserLightMode.ROTATING_COLORS)
            self.assertEqual(light.get_light_mode(), DiffuserLightMode.ROTATING_COLORS)
            await light.async_set_light_mode(mode=DiffuserLightMode.FIXED_RGB)
            self.assertEqual(light.get_light_mode(), DiffuserLightMode.FIXED_RGB)

    @unittest_run_loop
    async def test_light_brightness(self):
        if len(self.light_devices) < 1:
            self.skipTest("Could not find any DiffuserLightMixin within the given set of devices. "
                          "The test will be skipped.")
        for light in self.light_devices:
            print(f"Testing device: {light.name}...")
            await light.async_update()

            await light.async_set_light_mode(onoff=True)
            for i in range(0,100,10):
                await light.async_set_light_mode(brightness=i)
                await asyncio.sleep(0.5)
                self.assertEqual(light.get_light_brightness(), i)

    @unittest_run_loop
    async def test_spray_mode(self):
        if len(self.spray_devices) < 1:
            self.skipTest("Could not find any DiffuserSprayMixin within the given set of devices. "
                          "The test will be skipped.")
        for spray in self.spray_devices:  # type: Union[BaseDevice, DiffuserSprayMixin]
            print(f"Testing device: {spray.name}...")
            await spray.async_update()

            await asyncio.sleep(1)
            await spray.async_set_spray_mode(DiffuserSprayMode.LIGHT)
            self.assertEqual(spray.get_current_spray_mode(), DiffuserSprayMode.LIGHT)
            await asyncio.sleep(1)
            await spray.async_set_spray_mode(DiffuserSprayMode.STRONG)
            self.assertEqual(spray.get_current_spray_mode(), DiffuserSprayMode.STRONG)
            await asyncio.sleep(1)
            await spray.async_set_spray_mode(DiffuserSprayMode.OFF)
            self.assertEqual(spray.get_current_spray_mode(), DiffuserSprayMode.OFF)

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)



import asyncio
import json
from hashlib import md5
from typing import Dict
from uuid import uuid4
from threading import Event
from paho.mqtt import client as mqtt

from paho.mqtt.client import ssl, MQTTMessage

from meross_iot.utilities.mqtt import build_device_request_topic, build_client_response_topic, build_client_user_topic
from utilities.mixedqueue import MixedQueue


class AppSniffer(object):
    def __init__(self, logger, user_id, hashed_password, target_device_uuid, ca_cert=None, mqtt_host="iot.meross.com", mqtt_port=2001):
        self.l = logger
        self.connect_event = Event()
        self.subscribe_event = Event()
        self.user_id = user_id

        self.device_topic = None
        self.client_response_topic = None
        self.user_topic = None

        self.mqtt_domain = mqtt_host
        self.mqtt_port = mqtt_port
        self.target_device_uuid = target_device_uuid

        # Generate random app and client id
        md5_hash = md5()
        rnd_uuid = uuid4()
        md5_hash.update(("%s%s" % ("API", rnd_uuid)).encode("utf8"))
        self._app_id = "sniffer"
        self._client_id = 'app:sniffer-%s' % md5_hash.hexdigest()

        self._mqtt_client = mqtt.Client(client_id=self._client_id,
                                        protocol=mqtt.MQTTv311)

        self._mqtt_client.on_connect = self._on_connect
        self._mqtt_client.on_message = self._on_message
        self._mqtt_client.on_disconnect = self._on_disconnect
        self._mqtt_client.on_subscribe = self._on_subscribe

        # Avoid login if user_id is None
        if user_id is not None:
            self._mqtt_client.username_pw_set(username=user_id,
                                              password=hashed_password)
        self._mqtt_client.tls_set(ca_certs=ca_cert, certfile=None,
                                  keyfile=None, cert_reqs=ssl.CERT_REQUIRED,
                                  tls_version=ssl.PROTOCOL_TLS_CLIENT,
                                  ciphers=None)

        self._push_queue = MixedQueue(asyncio.get_running_loop())

    def _on_connect(self, client, userdata, rc, other):
        self.l.debug("Connected to MQTT Broker")
        self.connect_event.set()

    def start(self):
        """
        Starts the connection to the MQTT broker
        :return:
        """
        self.l.debug("Initializing the MQTT connection...")
        self._mqtt_client.connect(self.mqtt_domain, self.mqtt_port, keepalive=30)

        # Starts a new thread that handles mqtt protocol and calls us back via callbacks
        self.l.debug("(Re)Starting the MQTT loop.")
        self._mqtt_client.loop_stop(force=True)
        self._mqtt_client.loop_start()
        self.connect_event.wait()

        # Subscribe to the corresponding topics ...
        self.device_topic = build_device_request_topic(self.target_device_uuid)
        self.client_response_topic = build_client_response_topic(self.user_id, self._app_id)
        self.user_topic = build_client_user_topic(self.user_id)

        self.l.info(f"Subscribing to topic: {self.device_topic}")
        self._mqtt_client.subscribe(self.device_topic)
        self.subscribe_event.wait()
        self.subscribe_event.clear()

        self.l.info(f"Subscribing to topic: {self.client_response_topic}")
        self._mqtt_client.subscribe(self.client_response_topic)
        self.subscribe_event.wait()
        self.subscribe_event.clear()

        self.l.info(f"Subscribing to topic: {self.user_topic}")
        self._mqtt_client.subscribe(self.user_topic)
        self.subscribe_event.wait()
        self.subscribe_event.clear()

    def stop(self):
        self._mqtt_client.disconnect()
        self._mqtt_client.loop_stop(force=True)

    def _on_subscribe(self, client, userdata, mid, granted_qos):
        self.l.debug("Subscribed correctly")
        self.subscribe_event.set()

    def _on_message(self, client, userdata, msg):
        message = json.loads(str(msg.payload, "utf8"))
        header = message['header']

        topic_str = "Unknown"
        if msg.topic == self.user_topic:
            topic_str = "USER-TOPIC"
        elif msg.topic == self.client_response_topic:
            topic_str = "CLIENT-RESPONSE-TOPIC"
        elif msg.topic == self.device_topic:
            topic_str = "DEVICE-TOPIC"

        self.l.info("%s (%s) <- %s" % (topic_str, msg.topic, message))
        if header['method'].upper() == 'PUSH':
            self._push_queue.sync_put(message)

    def _on_disconnect(self, client, userdata, rc):
        self.l.debug("Disconnected from MQTT broker")

    async def async_wait_push_notification(self) -> Dict:
        return await self._push_queue.async_get()



import logging
import os

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.mixins.toggle import ToggleXMixin
from meross_iot.manager import MerossManager, TransportMode
from meross_iot.model.enums import OnlineStatus
from tests import async_get_client

if os.name == 'nt':
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


_LOGGER = logging.getLogger()
_LOGGER.setLevel(logging.DEBUG)


class TestTransportMode(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        devices = await self.meross_manager.async_device_discovery()
        toggle_devices = self.meross_manager.find_devices(online_status=OnlineStatus.ONLINE, device_class=ToggleXMixin)

        if len(toggle_devices) < 1:
            self.test_device = None
        else:
            self.test_device = toggle_devices[0]

    @unittest_run_loop
    async def test_update_default_transport_http_first(self):
        if self.test_device is None:
            self.skipTest("No device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")

        self.meross_manager.default_transport_mode = TransportMode.LAN_HTTP_FIRST

        for i in range(10):
            await self.test_device.async_update()
            await asyncio.sleep(1.0)

        self.meross_manager.default_transport_mode = TransportMode.MQTT_ONLY

    @unittest_run_loop
    async def test_update_default_transport_http_first_only_get(self):
        # Executes high rate updates via MQTT, and will probably fail
        if self.test_device is None:
            self.skipTest("No device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")

        self.meross_manager.default_transport_mode = TransportMode.LAN_HTTP_FIRST_ONLY_GET

        for i in range(10):
            await self.test_device.async_update()
            await asyncio.sleep(1.0)
            await self.test_device.async_toggle()

    @unittest_run_loop
    async def test_update_default_transport_http_first_with_errors(self):
        if self.test_device is None:
            self.skipTest("No device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")
        prev_lan_ip = self.test_device._inner_ip
        try:
            self.meross_manager.default_transport_mode = TransportMode.LAN_HTTP_FIRST

            # Set an invalid IP, this will make all HTTP request fail and be retried via MQTT
            self.test_device._inner_ip = "2.2.2.2"
            for i in range(10):
                await self.test_device.async_update()
                await asyncio.sleep(1.0)
                await self.test_device.async_toggle()
        except:
            self.test_device._inner_ip = prev_lan_ip
            raise

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)


import logging
from typing import Optional

from meross_iot.model.enums import Namespace, DiffuserSprayMode


_LOGGER = logging.getLogger(__name__)


class DiffuserSprayMixin(object):
    _execute_command: callable
    check_full_update_done: callable

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)

        # Dictionary keeping the status for every channel
        self._channel_diffuser_spray_status = {}

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False

        if namespace == Namespace.DIFFUSER_SPRAY:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace {namespace}")
            payload = data.get('spray')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'spray' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # Update the status of every channel that has been reported in this push
                # notification.
                for c in payload:
                    channel = c['channel']
                    self._channel_diffuser_spray_status[channel] = c

                locally_handled = True

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    async def async_handle_update(self, namespace: Namespace, data: dict) -> bool:
        _LOGGER.debug(f"Handling {self.__class__.__name__} mixin data update.")
        locally_handled = False
        if namespace == Namespace.SYSTEM_ALL:
            diffuser_data = data.get('all', {}).get('digest', {}).get('diffuser', {}).get('spray',[])
            for l in diffuser_data:
                channel = l['channel']
                self._channel_diffuser_spray_status[channel] = l
            locally_handled = True

        super_handled = await super().async_handle_update(namespace=namespace, data=data)
        return super_handled or locally_handled

    def get_current_spray_mode(self, channel: int = 0, *args, **kwargs) -> Optional[DiffuserSprayMode]:
        """
        Returns the current spray mode
        :param channel: channel to fetch info from
        :param args:
        :param kwargs:
        :return:
        """
        mode = self._channel_diffuser_spray_status.get(channel, {}).get('mode')
        if mode is not None:
            return DiffuserSprayMode(mode)
        return None

    async def async_set_spray_mode(self, mode: DiffuserSprayMode, channel: int = 0, timeout: Optional[float] = None, *args, **kwargs) -> None:
        """
        Changes the operating mode for this device
        :param mode: mode to set
        :param channel: channel to handle
        :param timeout: command timeout
        :return:
        """
        spray_payload = {"mode": mode.value, "channel": channel}
        payload = {'spray': [spray_payload]}
        await self._execute_command(method='SET',
                                    namespace=Namespace.DIFFUSER_SPRAY,
                                    payload=payload,
                                    timeout=timeout)
        # Immediately update local state
        self._channel_diffuser_spray_status[channel].update(spray_payload)



import os
from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.mixins.garage import GarageOpenerMixin
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus
from tests import async_get_client


if os.name == 'nt':
    import asyncio

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestGarageOpener(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        devices = await self.meross_manager.async_device_discovery()
        self.garage_devices = self.meross_manager.find_devices(device_class=GarageOpenerMixin, online_status=OnlineStatus.ONLINE, device_type="msg100")

    @unittest_run_loop
    async def test_open_close(self):
        if len(self.garage_devices) < 1:
            self.skipTest("Could not find any Garage Opener within the given set of devices. "
                          "The test will be skipped")

        garage = self.garage_devices[0]
        print(f"Testing device {garage.name}")

        # Without a full update, the status will be NONE
        current_status = garage.get_is_open()
        self.assertIsNone(current_status)

        # Trigger the full update
        await garage.async_update()
        self.assertIsNotNone(garage.get_is_open())

        # Toggle
        is_open = garage.get_is_open()
        if is_open:
            await garage.async_close()
        else:
            await garage.async_open()
        await asyncio.sleep(40)
        self.assertEqual(garage.get_is_open(), not is_open)

        is_open = garage.get_is_open()
        if is_open:
            await garage.async_close()
        else:
            await garage.async_open()
        await asyncio.sleep(40)
        self.assertEqual(garage.get_is_open(), not is_open)

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)



import os

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.http_api import MerossHttpClient
from meross_iot.model.http.error_codes import ErrorCodes
from meross_iot.model.http.exception import BadLoginException, BadDomainException, HttpApiError, MissingMFA, WrongMFA
from tests import async_get_client, _TEST_EMAIL, _TEST_PASSWORD, _TEST_API_BASE_URL, _TEST_EMAIL_MFA

if os.name == 'nt':
    import asyncio

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestHttpMethods(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

    @unittest_run_loop
    async def test_subdevice_listing(self):
        devices = await self.meross_client.async_list_devices()
        # look for a msxh0 hub
        hub = None
        for d in devices:
            if d.device_type == "msh300":
                hub = d
                break

        if hub is None:
            self.skipTest("No hub was found on this subscription. Cannot test hub listing.")
        print(f"Testing device {hub.dev_name}")
        result = await self.meross_client.async_list_hub_subdevices(hub.uuid)
        if len(result) < 1:
            self.skipTest("No subdevice found for selected hub")
            return
        self.assertGreater(len(result), 0)
        return result

    @unittest_run_loop
    async def test_bad_login(self):
        with self.assertRaises(BadLoginException):
            return await MerossHttpClient.async_from_user_password(api_base_url=_TEST_API_BASE_URL,
                                                                   email=_TEST_EMAIL,
                                                                   password="thisIzWRONG!")

    @unittest_run_loop
    async def test_not_existing_email(self):
        with self.assertRaises(HttpApiError):
            try:
                return await MerossHttpClient.async_from_user_password(api_base_url=_TEST_API_BASE_URL,
                                                                       email="thisDoesNotExistIGuess@gmail.com",
                                                                       password="thisIzWRONG!")
            except HttpApiError as e:
                self.assertEqual(e.error_code, ErrorCodes.CODE_WRONG_EMAIL)
                raise e

    @unittest_run_loop
    async def test_bad_email(self):
        with self.assertRaises(HttpApiError):
            try:
                return await MerossHttpClient.async_from_user_password(api_base_url=_TEST_API_BASE_URL,
                                                                       email="invalidemail",
                                                                       password="somePassword")
            except HttpApiError as e:
                self.assertEqual(e.error_code, ErrorCodes.CODE_WRONG_EMAIL)
                raise e

    @unittest_run_loop
    async def test_missing_mfa(self):
        with self.assertRaises(MissingMFA):
            return await MerossHttpClient.async_from_user_password(api_base_url=_TEST_API_BASE_URL,
                                                                   email=_TEST_EMAIL_MFA,
                                                                   password=_TEST_PASSWORD)

    @unittest_run_loop
    async def test_wrong_mfa(self):
        with self.assertRaises(WrongMFA):
            return await MerossHttpClient.async_from_user_password(api_base_url=_TEST_API_BASE_URL,
                                                                   email=_TEST_EMAIL_MFA,
                                                                   password=_TEST_PASSWORD,
                                                                   mfa_code="invalid")

    @unittest_run_loop
    async def test_device_listing(self):
        devices = await self.meross_client.async_list_devices()
        assert devices is not None
        assert len(devices) > 0


    @unittest_run_loop
    async def test_bad_domain(self):
        with self.assertRaises(BadDomainException):
            return await MerossHttpClient.async_from_user_password(api_base_url="iot.meross.com", email=_TEST_EMAIL, password=_TEST_PASSWORD, auto_retry_on_bad_domain=False)


    @unittest_run_loop
    async def test_subdevice_listing(self):
        devices = await self.meross_client.async_list_devices()
        # look for a msxh0 hub
        hub = None
        for d in devices:
            if d.device_type == "msh300":
                hub = d
                break

        if hub is None:
            self.skipTest("No hub was found on this subscription. Cannot test hub listing.")
        print(f"Testing device {hub.dev_name}")
        result = await self.meross_client.async_list_hub_subdevices(hub.uuid)
        self.assertGreater(len(result), 0)
        return result

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()



import logging
from hashlib import md5
import base64
from Crypto.Cipher import AES
from enum import Enum
from typing import Dict

from meross_iot.model.enums import Namespace

_LOGGER = logging.getLogger(__name__)


class EncryptionAlg(Enum):
    ECDHE256 = 0


class EncryptionSuiteMixin(object):
    _execute_command: callable
    _DEFAULT_IV="0000000000000000".encode("utf8")
    _abilities: Dict[str, dict]

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)
        self._encryption_key = None

        if Namespace.SYSTEM_ENCRYPTION_ECDHE.value in self._abilities:
            self._encryption_alg = EncryptionAlg.ECDHE256
        else:
            raise ValueError("Unsupported/undetected encryption method")

    def _pad_to_16_bytes(self, data):
        block_size = 16
        pad_length = block_size - (len(data) % block_size)
        padding = bytes([0] * pad_length)
        return data + padding

    def _ecdhe256_encrypt(self, message_data_bytes, iv=_DEFAULT_IV) -> str:
        # Returns encrypted message in base64 encoded string.
        padded_data = self._pad_to_16_bytes(message_data_bytes)
        cipher = AES.new(self._encryption_key, AES.MODE_CBC, iv)
        cipher.padding = 0
        encrypted = cipher.encrypt(padded_data)
        return base64.b64encode(encrypted).decode('utf-8')

    def _ecdhe256_decrypt(self, message_data_bytes: bytes, iv=_DEFAULT_IV) -> bytes:
        # Returns decrypted message bytes.
        cipher = AES.new(self._encryption_key, AES.MODE_CBC, iv)
        cipher.padding = 0
        enc_bytes = base64.b64decode(message_data_bytes)
        decrypted = cipher.decrypt(enc_bytes)
        return decrypted

    def support_encryption(self) -> bool:
        return True

    def is_encryption_key_set(self) -> bool:
        return self._encryption_key is not None

    def set_encryption_key(self, uuid:str, mrskey: str, mac: str, *args, **kwargs):
        strtohash = uuid[3:22] + mrskey[1:9] + mac + mrskey[10:28]
        self._encryption_key = md5(strtohash.encode("utf8")).hexdigest().encode("utf8")

    def encrypt(self, message_data_bytes: bytes)->str:
        """
        Encrypts the message into a base64 string
        :param message_data_bytes:
        :return:
        """
        if not self.is_encryption_key_set():
            raise ValueError("Encryption key is not set! Please invoke set_encryption_key first.")

        if self._encryption_alg == EncryptionAlg.ECDHE256:
            return self._ecdhe256_encrypt(message_data_bytes)

        raise ValueError("Unimplemented encryption algorithm")

    def decrypt(self, encrypted_message_bytes: bytes) -> bytes:
        """
        Decrypt the message and returns the war decrypted bytes
        :param encrypted_message_bytes:
        :return:
        """
        if not self.is_encryption_key_set():
            raise ValueError("Encryption key is not set! Please invoke set_encryption_key first.")

        if self._encryption_alg == EncryptionAlg.ECDHE256:
            return self._ecdhe256_decrypt(encrypted_message_bytes)

        raise ValueError("Unimplemented encryption algorithm")


import logging
from typing import Optional

from meross_iot.model.enums import Namespace, DiffuserLightMode
from meross_iot.model.typing import RgbTuple
from meross_iot.utilities.conversion import rgb_to_int, int_to_rgb

_LOGGER = logging.getLogger(__name__)


class DiffuserLightMixin(object):
    _execute_command: callable
    check_full_update_done: callable

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)

        # Dictionary keeping the status for every channel
        self._channel_diffuser_light_status = {}

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False

        if namespace == Namespace.DIFFUSER_LIGHT:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace {namespace}")
            payload = data.get('light')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'light' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # Update the status of every channel that has been reported in this push
                # notification.
                for c in payload:
                    channel = c['channel']
                    self._channel_diffuser_light_status[channel] = c

                locally_handled = True

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    def get_light_mode(self, channel: int = 0, *args, **kwargs) -> Optional[DiffuserLightMode]:
        """
        Returns the operating mode for this device
        :param channel: channel to fetch info from
        :return:
        """
        self.check_full_update_done()
        mode = self._channel_diffuser_light_status.get(channel, {}).get("mode")
        if mode is None:
            return None
        else:
            return DiffuserLightMode(mode)

    async def async_handle_update(self, namespace: Namespace, data: dict) -> bool:
        _LOGGER.debug(f"Handling {self.__class__.__name__} mixin data update.")
        locally_handled = False
        if namespace == Namespace.SYSTEM_ALL:
            diffuser_data = data.get('all', {}).get('digest', {}).get('diffuser', {}).get('light', [])
            for l in diffuser_data:
                channel = l['channel']
                self._channel_diffuser_light_status[channel] = l
            locally_handled = True

        super_handled = await super().async_handle_update(namespace=namespace, data=data)
        return super_handled or locally_handled

    def get_light_brightness(self, channel: int = 0, *args, **kwargs) -> Optional[int]:
        """
        Returns the current configured led brightness
        :param channel: channel index to fetch info from
        :return:
        """
        self.check_full_update_done()
        return self._channel_diffuser_light_status.get(channel, {}).get("luminance")

    def get_light_rgb_color(self, channel=0, *args, **kwargs) -> Optional[RgbTuple]:
        """
        Returns the current RGB configuration of the device.

        :param channel: channel to control, defaults to 0.

        :return: a Tuple containing three integer 8bits values (red, green, blue)
        """
        self.check_full_update_done()
        info = self._channel_diffuser_light_status.get(channel, {}).get('rgb')
        if info is None:
            return None
        return int_to_rgb(info)

    async def async_set_light_mode(self, channel: int = 0, onoff: bool = None, mode: DiffuserLightMode = None,
                                   brightness: int = None, rgb: Optional[RgbTuple] = None,
                                   timeout: Optional[float] = None, *args, **kwargs) -> None:
        """
        Sets the light mode for this device.
        :param channel: channel to configure
        :param onoff: when True, sets the light ON, otherwise OFF.
        :param mode: defines the operation mode for this light
        :param brightness: brightness value (from 0 to 100)
        :param rgb: tuple of three integers (each from 0 to 255) for red, green, blue
        :param timeout: command timeout
        :return:
        """
        light_payload = {'channel': channel}
        if mode is not None:
            light_payload['mode'] = mode.value
        if brightness is not None:
            light_payload['luminance'] = brightness
        if rgb is not None:
            light_payload['rgb'] = rgb_to_int(rgb)
        if onoff is not None:
            light_payload['onoff'] = 1 if onoff else 0
        payload = {'light': [light_payload]}
        await self._execute_command(method='SET',
                                    namespace=Namespace.DIFFUSER_LIGHT,
                                    payload=payload,
                                    timeout=timeout)
        # Immediately update local state
        if channel not in self._channel_diffuser_light_status:
            self._channel_diffuser_light_status[channel] = {}

        self._channel_diffuser_light_status[channel].update(light_payload)

    def get_light_is_on(self, channel: int = 0, *args, **kwargs) -> Optional[bool]:
        """
        Returns True if the light is ON, False otherwise.

        :param channel: channel to control, defaults to 0 (bulbs generally have only one channel)

        :return: current onoff state
        """
        self.check_full_update_done()
        onoff = self._channel_diffuser_light_status.get(channel, {}).get('onoff')
        if onoff is None:
            return None
        return onoff == 1

    async def async_turn_on(self, channel=0, *args, **kwargs) -> None:
        await self.async_set_light_mode(channel=channel, onoff=True)

    async def async_turn_off(self, channel=0, *args, **kwargs) -> None:
        await self.async_set_light_mode(channel=channel, onoff=False)



import asyncio


class MixedQueue:
    def __init__(self, loop):
        self._queue = asyncio.Queue()
        self._loop = loop

    def sync_put_nowait(self, item):
        self._loop.call_soon(self._queue.put_nowait, item)

    def sync_put(self, item):
        asyncio.run_coroutine_threadsafe(self._queue.put(item), self._loop).result()

    def sync_get(self):
        return asyncio.run_coroutine_threadsafe(self._queue.get(), self._loop).result()

    def async_put_nowait(self, item):
        self._queue.put_nowait(item)

    async def async_put(self, item):
        await self._queue.put(item)

    async def async_get(self):
        return await self._queue.get()